require "fluent/plugin/output"
require "net/http/persistent"

module Fluent::Plugin
  class ContainiqOutput < Fluent::Plugin::Output
    Fluent::Plugin.register_output("containiq", self)

    config_param :api_key, :string, default: nil, secret: true
    config_param :bulk_limit, :integer, default: 1000000 # Logz.io has a 1MB limit and recommends leaving some overhead
    config_param :bulk_limit_warning_limit, :integer, default: nil # If fluent warnings are sent to the output, truncating is necessary to prevent a recursion
    config_param :http_idle_timeout, :integer, default: 5
    config_param :gzip, :bool, default: true

    def configure(conf)
      super

      # ensure endpoint url is configured correctly, otherwise raise an error
      get_endpoint_url()
      get_api_key(@api_key)
    end

    def start
      super

      endpoint_url = get_endpoint_url()
      @uri = URI endpoint_url
      log.debug "ContainIQ URL #{endpoint_url}"

      @http = Net::HTTP::Persistent.new name: 'fluent-plugin-containiq'

      api_key = get_api_key(@api_key)
      @http.headers['Authorization'] = "Bearer #{api_key}"

      @http.headers['Content-Type'] = 'text/plain'
      if @gzip
        @http.headers['Content-Encoding'] = 'gzip'
      end
      @http.idle_timeout = @http_idle_timeout
      @http.socket_options << [Socket::SOL_SOCKET, Socket::SO_KEEPALIVE, 1]

      log.debug "Started ContainIQ shipper.."
    end

    def shutdown
      super
    end

    def formatted_to_msgpack_binary?
      true
    end

    def multi_workers_ready?
      true
    end

    def format(tag, time, record)
      if time.is_a?(Fluent::EventTime)
        sec_frac = time.to_f
      else
        sec_frac = time * 1.0
      end
      [tag, sec_frac, record].to_msgpack
    end

    def write(chunk)
      encode_chunk(chunk) { |bulk_records, bulk_size|
        send_bulk(bulk_records, bulk_size)
      }
    end

    def encode_chunk(chunk)
      records = []
      bulk_size = 0
      cluster_name = ENV["CLUSTER_NAME"]
      if cluster_name.nil?
        cluster_name = "default"
      end
      chunk.each { |tag, time, record|
        record['timestamp'] ||= Time.at(time).iso8601(3)
        record['cluster'] ||= cluster_name

        begin
          json_record = Yajl.dump(record)
          record_size = json_record.size + (1 if !records.empty?).to_i # Accounting for trailing "\n"
        rescue
          log.error "Adding record #{record} to buffer failed. Exception: #{$!}"
          next
        end

        if record_size > @bulk_limit
          if @bulk_limit_warning_limit.is_a?(Integer)
            log.warn "Record with size #{record_size} exceeds #{@bulk_limit} and can't be sent to ContainIQ. Record starts with (truncated at #{@bulk_limit_warning_limit} characters): #{json_record[0,@bulk_limit_warning_limit]}"
            # Send the full message to debug facility
            log.debug "Record with size #{record_size} exceeds #{@bulk_limit} and can't be sent to ContainIQ. Record is: #{json_record}"
          else
            log.warn "Record with size #{record_size} exceeds #{@bulk_limit} and can't be sent to ContainIQ. Record is: #{json_record}"
          end
          next
        end
        if bulk_size + record_size > @bulk_limit
          yield(records, bulk_size)
          records = []
          bulk_size = 0
        end
        records.push(json_record)
        bulk_size += record_size
      }
      if records
        yield(records, bulk_size)
      end
    end

    def send_bulk(bulk_records, bulk_size)
      log.debug "Sending a bulk of #{bulk_records.size} records, size #{bulk_size}B to ContainIQ"

      # Setting our request
      post = Net::HTTP::Post.new @uri.request_uri

      body = "[#{bulk_records.join(",")}]"
      if gzip
        post.body = compress(body)
      else
        post.body = body
      end

      retry_count = 4     # How many times to resend failed bulks
      sleep_interval = 2  # How long to sleep initially between retries

      begin
        retry_count.times do |counter|
          should_retry = true
          begin
            response = @http.request @uri, post
            if response.code != '200'
              if response.code == '401'
                log.error "You are not authorized with ContainIQ! Token OK? dropping logs..."
                should_retry = false
              elsif response.code == '400'
                log.info "Got 400 code from ContainIQ. This means that some of your logs are too big, or badly formatted. Response: #{response.body}"
                should_retry = false
              else
                log.warn "Got HTTP #{response.code} from ContainIQ, not giving up just yet (Try #{counter + 1}/#{retry_count})"
              end
            else
              log.debug "Successfully sent bulk of #{bulk_records.size} records, size #{bulk_size}B to ContainIQ"
              should_retry = false
            end
          rescue StandardError => e
            log.warn "Error connecting to ContainIQ. Got exception: #{e} (Try #{counter + 1}/#{retry_count})"
          end

          if should_retry
            if counter == retry_count - 1
              log.error "Could not send your bulk after #{retry_count} tries Sorry! Your bulk is: #{post.body}"
              break
            end
            sleep(sleep_interval)
            sleep_interval *= 2
          else
            return
          end
        end
      rescue Exception => e
        log.error "Got unexpected exception! Here: #{e}"
      end
    end

    def compress(string)
      wio = StringIO.new("w")
      w_gz = Zlib::GzipWriter.new(wio)
      w_gz.write(string)
      w_gz.close
      wio.string
    end

    def get_endpoint_url
      endpoint_url = ENV["INGEST_LOGS_ENDPOINT_URL"]
      raise 'missing environment variable: INGEST_LOGS_ENDPOINT_URL' if endpoint_url.nil?
      endpoint_url
    end

    def get_api_key(api_key)
      # if api key is null, get it from a local file called "NOTIFICATION_FILE_LOCATION"
      if api_key.nil?
        file = ENV["NOTIFICATION_FILE_LOCATION"]
        raise Fluent::ConfigError, 'missing environment variable: NOTIFICATION_FILE_LOCATION' if file.nil?
        fileFirstLine = File.open(file, &:readline)
        scan = fileFirstLine.gsub("\n",'').scan(/key: (.+)/i)
        raise Fluent::ConfigError, 'unable to parse secret key' if scan.empty?
        api_key = scan.first.first
      end
      return api_key
    end
  end
end
